﻿using System;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;

namespace TestLibrary
{
    /// <summary>
    /// 이미지 색상 선택기
    /// </summary>
    public class ImageColorPicker : Image
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Dependency Property
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 셀렉터 속성 - SelectorProperty

        /// <summary>
        /// 셀렉터 속성
        /// </summary>
        public static readonly DependencyProperty SelectorProperty = DependencyProperty.Register
        (
            "Selector",
            typeof(Drawing),
            typeof(ImageColorPicker),
            new FrameworkPropertyMetadata
            (
                new GeometryDrawing(Brushes.White, new Pen(Brushes.Black, 1),
                new EllipseGeometry(new Point(), 5, 5)),
                FrameworkPropertyMetadataOptions.AffectsRender
            ),
            ValidateSelectorPropertyValueCallback
        );

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Dependency Property Key
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Private

        #region 선택 색상 속성 키 - SelectedColorPropertyKey

        /// <summary>
        /// 선택 색상 속성 키
        /// </summary>
        private static readonly DependencyPropertyKey SelectedColorPropertyKey = DependencyProperty.RegisterReadOnly
        (
            "SelectedColor",
            typeof(Color), typeof(ImageColorPicker),
            new FrameworkPropertyMetadata
            (
                Colors.Transparent,
                FrameworkPropertyMetadataOptions.AffectsRender
            )
        );

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 타겟 비트맵
        /// </summary>
        private RenderTargetBitmap targetBitmap;

        /// <summary>
        /// 위치
        /// </summary>
        private Point position = new Point();

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 셀렉터 - Selector

        /// <summary>
        /// 셀렉터
        /// </summary>
        public Drawing Selector
        {
            get
            {
                return (Drawing)GetValue(SelectorProperty);
            }
            set
            {
                SetValue(SelectorProperty, value);
            }
        }

        #endregion
        #region 위치 - Position

        /// <summary>
        /// 위치
        /// </summary>
        Point Position
        {
            get
            {
                return this.position;
            }
            set
            {
                Point newPosition = GetPosition(value);

                if(this.position != newPosition)
                {
                    this.position = newPosition;

                    Color color = GetColor(this.position.X, this.position.Y);

                    if(color == SelectedColor)
                    {
                        InvalidateVisual();
                    }

                    SetValue(SelectedColorPropertyKey, color);
                }
            }
        }

        #endregion
        #region 선택 색상 - SelectedColor

        /// <summary>
        /// 선택 색상
        /// </summary>
        public Color SelectedColor
        {
            get
            {
                return (Color)GetValue(SelectedColorPropertyKey.DependencyProperty);
            }
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 타겟 비트맵

        /// <summary>
        /// 타겟 비트맵
        /// </summary>
        private RenderTargetBitmap TargetBitmap
        {
            get 
            {
                if(this.targetBitmap == null)
                {
                    DrawingImage drawingImage = Source as DrawingImage;

                    if(drawingImage != null)
                    {
                        DrawingVisual drawingVisual = new DrawingVisual();

                        using(DrawingContext drawingContext = drawingVisual.RenderOpen())
                        {
                            drawingContext.DrawDrawing(drawingImage.Drawing);
                        }

                        Rect rectangle = drawingVisual.ContentBounds;

                        drawingVisual.Transform = new ScaleTransform
                        (
                            ActualWidth  / rectangle.Width,
                            ActualHeight / rectangle.Height
                        );

                        this.targetBitmap = new RenderTargetBitmap
                        (
                            (int)ActualWidth,
                            (int)ActualHeight,
                            96,
                            96,
                            PixelFormats.Pbgra32
                        );

                        this.targetBitmap.Render(drawingVisual);
                    }
                }

                return this.targetBitmap; 
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Private

        #region 셀렉터 속성 값 검증시 콜백 처리하기 - ValidateSelectorPropertyValueCallback(value)

        /// <summary>
        /// 셀렉터 속성 값 검증시 콜백 처리하기
        /// </summary>
        /// <param name="value">값</param>
        /// <returns>셀렉터 속성 값 검증 결과</returns>
        private static bool ValidateSelectorPropertyValueCallback(object value)
        {
            return value == null ? false : true;
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Protected

        #region 속성 변경시 처리하기 - OnPropertyChanged(e)

        /// <summary>
        /// 속성 변경시 처리하기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected override void OnPropertyChanged(DependencyPropertyChangedEventArgs e)
        {
            if(e.Property.Name == "Source")
            {
                this.targetBitmap = null;

                Position = new Point();
            }

            base.OnPropertyChanged(e);
        }

        #endregion
        #region 렌더링 크기 변경시 처리하기 - OnRenderSizeChanged(info)

        /// <summary>
        /// 렌더링 크기 변경시 처리하기
        /// </summary>
        /// <param name="info">크기 변경시 정보</param>
        protected override void OnRenderSizeChanged(SizeChangedInfo info)
        {
            base.OnRenderSizeChanged(info);

            this.targetBitmap = null;

            if(info.PreviousSize.Width > 0 && info.PreviousSize.Height > 0)
            {
                Position = new Point
                (
                    Position.X * info.NewSize.Width  / info.PreviousSize.Width,
                    Position.Y * info.NewSize.Height / info.PreviousSize.Height
                );
            }
        }

        #endregion
        #region 렌더링시 처리하기 - OnRender(context)

        /// <summary>
        /// 렌더링시 처리하기
        /// </summary>
        /// <param name="context">드로잉 컨텍스트</param>
        protected override void OnRender(DrawingContext context)
        {
            base.OnRender(context);

            if(ActualWidth == 0 || ActualHeight == 0)
            {
                return;
            }

            context.PushTransform(new TranslateTransform(Position.X, Position.Y));

            context.DrawDrawing(Selector);

            context.Pop();
        }

        #endregion

        #region 마우스 진입시 처리하기 - OnMouseEnter(e)

        /// <summary>
        /// 마우스 진입시 처리하기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected override void OnMouseEnter(MouseEventArgs e)
        {
            base.OnMouseEnter(e);

            if(e.LeftButton == MouseButtonState.Pressed)
            {
                Point mousePoint = e.GetPosition(this);

                Position = new Point(mousePoint.X, mousePoint.Y);
            }
        }

        #endregion
        #region 마우스 왼쪽 버튼 DOWN 처리하기 - OnMouseLeftButtonDown(e)

        /// <summary>
        /// 마우스 왼쪽 버튼 DOWN 처리하기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected override void OnMouseLeftButtonDown(MouseButtonEventArgs e)
        {
            base.OnMouseLeftButtonDown(e);

            SetPosition(e.GetPosition(this));
        }

        #endregion
        #region 마우스 이동시 처리하기 - OnMouseMove(e)

        /// <summary>
        /// 마우스 이동시 처리하기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected override void OnMouseMove(MouseEventArgs e)
        {
            base.OnMouseMove(e);

            if(e.LeftButton == MouseButtonState.Pressed)
            {
                SetPosition(e.GetPosition(this));
            }
        }

        #endregion
        #region 마우스 왼쪽 버튼 UP 처리하기 - OnMouseLeftButtonUp(e)

        /// <summary>
        /// 마우스 왼쪽 버튼 UP 처리하기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected override void OnMouseLeftButtonUp(MouseButtonEventArgs e)
        {
            base.OnMouseLeftButtonUp(e);

            SetPosition(e.GetPosition(this));
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Private

        #region 위치 구하기 - GetPosition(sourcePoint)

        /// <summary>
        /// 위치 구하기
        /// </summary>
        /// <param name="sourcePoint">소스 위치</param>
        /// <returns>위치</returns>
        private Point GetPosition(Point sourcePoint)
        {
            double x = sourcePoint.X;
            double y = sourcePoint.Y;

            if(x < 0)
            {
                x = 0;
            }
            else if(x > ActualWidth)
            {
                x = ActualWidth;
            }
            
            if(y < 0)
            {
                y = 0;
            }
            else if(y > ActualHeight)
            {
                y = ActualHeight;
            }
            
            return new Point(x, y);
        }

        #endregion
        #region 위치 설정하기 - SetPosition(sourcePoint)

        /// <summary>
        /// 위치 설정하기
        /// </summary>
        /// <param name="sourcePoint">소스 포인트</param>
        private void SetPosition(Point sourcePoint)
        {
            if(sourcePoint.X >= 0 && sourcePoint.X <= ActualWidth && sourcePoint.Y >= 0 && sourcePoint.Y <= ActualHeight)
            {
                Position = sourcePoint;
            }
        }

        #endregion
        #region 색상 구하기 - GetColor(x, y)

        /// <summary>
        /// 색상 구하기
        /// </summary>
        /// <param name="x">X</param>
        /// <param name="y">Y</param>
        /// <returns>색상</returns>
        private Color GetColor(double x, double y)
        {
            if(Source == null)
            {
                throw new InvalidOperationException("Image Source not set");
            }
            
            BitmapSource bitmapSource = Source as BitmapSource;

            if(bitmapSource != null)
            {
                x *= bitmapSource.PixelWidth / ActualWidth;

                if((int)x > bitmapSource.PixelWidth - 1)
                {
                    x = bitmapSource.PixelWidth - 1;
                }
                else if(x < 0)
                {
                    x = 0;
                }

                y *= bitmapSource.PixelHeight / ActualHeight;

                if((int)y > bitmapSource.PixelHeight - 1)
                {
                    y = bitmapSource.PixelHeight - 1;
                }
                else if(y < 0)
                {
                    y = 0;
                }

                if(bitmapSource.Format == PixelFormats.Indexed4)
                {
                    byte[] pixelByteArray = new byte[1];

                    int stride = (bitmapSource.PixelWidth * bitmapSource.Format.BitsPerPixel + 3) / 4;

                    bitmapSource.CopyPixels(new Int32Rect((int)x, (int)y, 1, 1), pixelByteArray, stride, 0);

                    return bitmapSource.Palette.Colors[pixelByteArray[0] >> 4];
                }
                else if(bitmapSource.Format == PixelFormats.Indexed8)
                {
                    byte[] pixelByteArray = new byte[1];

                    int stride = (bitmapSource.PixelWidth * bitmapSource.Format.BitsPerPixel + 7) / 8;

                    bitmapSource.CopyPixels(new Int32Rect((int)x, (int)y, 1, 1), pixelByteArray, stride, 0);

                    return bitmapSource.Palette.Colors[pixelByteArray[0]];
                }
                else
                {
                    byte[] pixelByteArray = new byte[4];

                    int stride = (bitmapSource.PixelWidth * bitmapSource.Format.BitsPerPixel + 7) / 8;

                    bitmapSource.CopyPixels(new Int32Rect((int)x, (int)y, 1, 1), pixelByteArray, stride, 0);

                    return Color.FromArgb(pixelByteArray[3], pixelByteArray[2], pixelByteArray[1], pixelByteArray[0]);
                }
            }

            DrawingImage drawingImage = Source as DrawingImage;

            if(drawingImage != null)
            {
                RenderTargetBitmap renderTargetBitmap = TargetBitmap;

                x *= renderTargetBitmap.PixelWidth / ActualWidth;

                if((int)x > renderTargetBitmap.PixelWidth - 1)
                {
                    x = renderTargetBitmap.PixelWidth - 1;
                }
                else if(x < 0)
                {
                    x = 0;
                }

                y *= renderTargetBitmap.PixelHeight / ActualHeight;

                if((int)y > renderTargetBitmap.PixelHeight - 1)
                {
                    y = renderTargetBitmap.PixelHeight - 1;
                }
                else if(y < 0)
                {
                    y = 0;
                }

                byte[] pixelByteArray = new byte[4];

                int stride = (renderTargetBitmap.PixelWidth * renderTargetBitmap.Format.BitsPerPixel + 7) / 8;

                renderTargetBitmap.CopyPixels(new Int32Rect((int)x, (int)y, 1, 1), pixelByteArray, stride, 0);

                return Color.FromArgb(pixelByteArray[3], pixelByteArray[2], pixelByteArray[1], pixelByteArray[0]);
            }

            throw new InvalidOperationException("Unsupported Image Source Type");
        }

        #endregion
    }
}